/**
dungsurv.c (Large model)
Program to convert estimate of elephant dung pile density to estimate of
elephant density
28 Apr 92
**/

#include <alloc.h>
#include <conio.h>
#include <graphics.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <datetime.h>
#include <extraio.h>
#include <handy.h>
#include <iscreen.h>
#include <menu.h>
#include <plot.h>
#include <readkey.h>
#include <seeplot.h>
#include <stringx.h>

#define DCFG_FILE "DUNGSURV.DFT"
#define DEFDATFILE "EXAMPLE.DAT"
#define DEFDIR ""
#define DEFPROD 15.0
#define DEFSTAGES 5
#define NEXP 250
#define NSTAGES 10
#define N_PLOT    2
#define N_UNITS   6
#define PLOTFILE   "REMAIN.PLT"
#define VER_TEXT "Version 2.12"
#define WEIGHTFILE "WEIGHTS.PLT"

#define FCOL WHITE
#define BCOL BLUE
#define HICOL YELLOW
#define INP_FCOL BLACK
#define INP_BCOL LIGHTGRAY
#define FLAGCOL LIGHTRED
#define MENU_FCOL MK_ATTR(LIGHTGRAY,BLACK)
#define MENU_BCOL MK_ATTR(BLACK,LIGHTGRAY)
#define MENU_HICOL MENU_BCOL

static int comp_date(const void *a,const void *b);
static unsigned days(const char *s,int *day,int *month,int *year);
static char *list_stages(char *s,int upto);
static void newmsg(char text[],char highlight);
static void quit(void);
static char *read_line(char *s);

typedef struct {
	double prop[NSTAGES],production;
	unsigned dat,n_original;
	} experiment_data;

static experiment_data e[NEXP];
static FILE *input,*dcfg=NULL;
static double ele_density[NSTAGES];
static int nstages;
static BYTE scrnbuff[4000];
static char stname[NSTAGES][3]={"A","B","C1","C2","D","E"};

main()
{
static char *intext[2]={"Elephant dung survey","analysis"},
			*choices[2]={" Interactive ","  From file  "},
			*samedate[2]={"Re-enter data for this date",
						  "      Change the date      "},
			*units_list[N_UNITS]=
	{"         dung piles         ",
	 "     dung piles per km\xFD     ",
	 "   dung piles per hectare   ",
	 "      dung piles per m\xFD     ",
	 "    dung piles per acre     ",
	 " dung piles per square yard "},
		   *plot_choices[N_PLOT]={" % dung piles surviving to survey date ",
								  "          Weighting function           "};
MENU_TYPE m;
FILE *output;
long lastline;
static double cv[NSTAGES],*dt,f[NSTAGES],f_without_prod,maxwt,
			  pile_density[NSTAGES],surv_prod=DEFPROD,tdiff,temp,
			  units_conversion[N_UNITS] ={1.0,1.0,100.0,1.0e6,247.105381,
			  1.195990046301e6},wfact,var,xstep,*y,ystep;
int day,month,year,sday,smonth,syear,expno,i,j,line,ne,original,ret,s,
	sum,surviving[NSTAGES],units,wdp;
unsigned surv_date,exdate;
char add_on,*cp,gfx,str[256],filename[80]=DEFDATFILE,warnings;

gfx = initial_screen(intext,2,VER_TEXT,NULL,DEFDIR,NULL,NULL,TRUE);
getch();
restorecrtmode();
window(1,3,80,25);
textcolor(FCOL);
textbackground(BCOL);
clrscr();
newmsg("Data can be entered interactively at the keyboard or read in from a"
	   " file\n\r Please choose a type of data-entry",FALSE);
if ((ret = menu(5,4,1,2,choices,"Data-entry",&m,MENU_FCOL,MENU_BCOL,
				MENU_HICOL,NULL)) == -1)
		quit();
if ((dcfg = fopen(DCFG_FILE,"r")) != NULL)
	{
	strip(fgets(str,80,dcfg),"[]");
	if (cp = strchr(str,'\n'))
		*cp = '\0';
	strcpy(filename,strfirst(str));
	sscanf(strip(fgets(str,80,dcfg),"[]"),"%lf",&surv_prod);
	}
if (ret == 0)
	{
	input = stdin;
	newmsg("Entries made in this session will be logged in a file\n\r"
		   "Please enter a name for this file (or press [Esc] to quit)",FALSE);
	puts_at(4,10,"Log file:");
	if (xgets(filename,12,15,12,79,64,xgetch,INP_FCOL,INP_BCOL) == NULL)
		quit();
	while ((output = fopen(strupr(filename),"w"))== NULL)
		{
		sprintf(str,"Error: Cannot open the file %s\n\rPlease change the name"
		" (or press [Esc] to quit)",filename);
		newmsg(str,TRUE);
		if (xgets(filename,12,15,12,79,64,xgetch,INP_FCOL,INP_BCOL) == NULL)
			quit();
		}
	}
else
	{
	output = stdout;
	newmsg("Enter the name of a file which contains data (or press [Esc] to"
		   " quit)",FALSE);
	puts_at(4,10,"Data file:");
	if (xgets(filename,12,15,12,79,64,xgetch,INP_FCOL,INP_BCOL) == NULL)
		quit();
	while ((input = fopen(strupr(filename),"r+"))== NULL)
		{
		sprintf(str,"Error: Cannot open the file %s\n\rPlease change the name"
		" (or press [Esc] to quit)",filename);
		newmsg(str,TRUE);
		if (xgets(filename,12,15,12,79,64,xgetch,INP_FCOL,INP_BCOL) == NULL)
			quit();
		}
	}
clearmenu(&m);
clrscr();
newmsg("Enter date of survey (dd/mm/yy): ",FALSE);
cputs("Date of survey: ");
read_line(str);
while ((surv_date = days(str,&sday,&smonth,&syear)) == 0)
	{
	newmsg("Invalid date - please re-enter (dd/mm/yy)",TRUE);
	cputs("Date of survey: ");
	read_line(str);
	}
if (input == stdin)
	fprintf(output,"[Date of survey] %d/%d/%d\n",sday,smonth,syear);
else
	{
	textcolor(HICOL);
	cputs(strcat(str,"\r"));
	textcolor(FCOL);
	}
sprintf(str,"Enter the number of stages used to classify dung\n\r (press "
		"[Enter] to accept the default %d stages)",DEFSTAGES);
newmsg(str,FALSE);
cputs("Number of stages: ");
read_line(str);
nstages = DEFSTAGES;
sscanf(str,"%d",&nstages);
while (nstages < 1 || nstages > NSTAGES)
	{
	sprintf(str,"Error: The number of stages must be between 1 and %d\n\r"
			"(default is %d stages)",NSTAGES,DEFSTAGES);
	newmsg(str,TRUE);
	cputs("Number of stages: ");
	read_line(str);
	nstages = DEFSTAGES;
	sscanf(str,"%d",&nstages);
	}
for (i = 0; i < nstages; ++i)
	{
	pile_density[i] = -1.0;
	sscanf(strip(fgets(str,80,dcfg),"[]"),"%lf",&pile_density[i]);
	}
fclose(dcfg);
if (input == stdin)
	fprintf(output,"[Number of stages] %d\n",nstages);
else
	{
	textcolor(HICOL);
	cprintf("%d\n\r",nstages);
	textcolor(FCOL);
	}
if (nstages < DEFSTAGES)
	for (i = 0; i < nstages; ++i)
		{
		stname[i][0] = '1' + i;
		stname[i][1] = '\0';
		}
if (nstages > 1)
	{
	cprintf("\n\rThe %d stages will be referred to as %s",nstages,stname[0]);
	for (i = 1; i < nstages - 1; ++i)
		cprintf(", %s",stname[i]);
	cprintf(" and %s.\n\r",stname[nstages - 1]);
	}
ne = 0;
e[0].production = DEFPROD;
newmsg("Experiment 1: Enter the date that fresh dung piles were observed\n\r"
	   " (dd/mm/yy)",FALSE);
cputs("\n\rDate           : ");
str[0] = '\0';
lastline = ftell(input);
do
	{
	while (ne < NEXP && read_line(str)[0] != '*')
		{
		while (str[0] != '*' && ((exdate = days(str,&day,&month,&year)) == 0
		  || exdate > surv_date))
			{
			newmsg((exdate == 0) ? "Invalid date - please re-enter in the form "
				   "dd/mm/yy" : "Experiment must be before survey!\n\r Please"
				   " re-enter date (dd/mm/yy)",TRUE);
			cputs("\n\rDate           : ");
			read_line(str);
			}
		if (str[0] == '*')
			break;
		expno = ne;
		for (i = 0; i < ne && e[i].dat != exdate; ++i);
		if (i < ne)
			{
			if (input == stdin)
				{
				sprintf(str,"Data has already been entered for %d/%d/%d\n\r"
						"Would you like to re-enter this data or enter a new"
						" date?",day,month,year);
				newmsg(str,TRUE);
				ret = menu(5,4,1,2,samedate,"Existing date",&m,MENU_FCOL,
						   MENU_BCOL,MENU_HICOL,NULL);
				clearmenu(&m);
				if (ret == -1)
					break;
				if (ret == 0)
					expno = i;
				else
					{
					sprintf(str,"Experiment %d: Enter the date that fresh dung piles"
							" were observed (dd/mm/yy)\n\r put * when finished",ne + 1);
					newmsg(str,FALSE);
					cputs("\n\rDate           : ");
					lastline = ftell(input);
					continue;
					}
				}
			else
				expno = i;
			}
		if (input != stdin)
			cputs(strcat(str,"\r"));
		newmsg("Enter the number of fresh dung piles noted for this experiment",
			   FALSE);
		cputs("Original number: ");
		original = 0;
		sscanf(read_line(str),"%d",&original);
		if (original <= 0)
			{
			sprintf(str,"Entries for experiment %d abandoned: Enter a new date"
			" that fresh\n\r dung piles were observed (dd/mm/yy) or put * if"
			" finished",expno + 1);
			newmsg(str,TRUE);
			cputs("\n\rDate           : ");
			continue;
			}
		if (input == stdin)
			fprintf(output,"[Date of experiment %d] %u/%u/%u\n[Number of dung"
					" piles noted] %d\n",expno + 1,day,month,year,original);
		else
			cputs(strcat(str,"\r"));
		for (i = 0,sum = 0; i < nstages; ++i)
			{
			sprintf(str,"Enter the number of piles which were in stage %s at"
					" the time of the survey",stname[i]);
			newmsg(str,FALSE);
			cprintf(" Stage %-2s  : ",stname[i]);
			surviving[i] = 0;
			sscanf(read_line(str),"%d",&surviving[i]);
			while (surviving[i] < 0 || surviving[i] + sum > original)
				{
				newmsg("Illegal entry - please re-enter the number of the"
					   " original piles which\n\rwere in this stage at the"
					   " time of the survey",TRUE);
				cprintf(" Stage %-2s  : ",stname[i]);
				sscanf(read_line(str),"%d",&surviving[i]);
				}
			sum += surviving[i];
			if (input == stdin)
				fprintf(output,"[# piles in stage %s] %d\n",stname[i],
						surviving[i]);
			else
				cputs(strcat(str,"\r"));
			if (i > 0)
				surviving[i] += surviving[i - 1];
			}
		if (expno > 0)
			e[expno].production = e[expno - 1].production;
		sprintf(str,"Enter the rate of dung pile production at %u/%u/%u in"
				" dung piles per elephant\n\rper day (default is %g dung"
				" piles per elephant per day)",day,month,year,
				e[expno].production);
		newmsg(str,FALSE);
		cputs("Defecation rate: ");
		sscanf(read_line(str),"%lf",&e[expno].production);
		if (input == stdin)
			fprintf(output,"[Defecation rate] %g\n",e[expno].production);
		else
			cputs(strcat(str,"\r"));
		for (i = 0; i < nstages; ++i)
			e[expno].prop[i] = 100.0*surviving[i]/original;
		e[expno].n_original = original;
		e[expno].dat = exdate;
		sprintf(str,"Experiment %d: Enter the date that fresh dung piles"
				" were observed (dd/mm/yy)\n\r put * when finished",ne + 2);
		newmsg(str,FALSE);
		cputs("\n\rDate           : ");
		str[0] = '\0';
		lastline = ftell(input);
		if (expno == ne)
			++ne;
		}
	if (input == stdin)
		{
		fprintf(output,"[End of experiment data]*\n");
		if (ne == 0)
			{
			fclose(output);
			quit();
			}
		}
	else
		cputs("\n\r");
	if (input == stdin)
		{
		fclose(output);
		add_on = FALSE;
		}
	else
		{
		output = input;
		input = stdin;
		fseek(output,lastline,SEEK_SET);
		printf("\n\n*** Details of %d experiments have been read from %s ***\n",
			   ne,filename);
		newmsg("Enter the start date of another experiment (dd/mm/yy)\n\r"
			   "or put * if finished",FALSE);
		cputs("\n\rDate           : ");
		add_on = TRUE;
		}
	}
while (add_on);
e[ne].production = surv_prod;
newmsg("Enter the rate of dung pile production at the time of the survey\n\r"
	   " (in dung piles per elephant per day)",FALSE);
cputs("\n\rDefecation rate at survey: ");
sprintf(str,"%g",surv_prod);
xgets(str,wherey() + 2,30,wherey() + 2,77,47,xgetch,INP_FCOL,INP_BCOL);
if ((temp = atof(str)) > 0.0)
	surv_prod = e[ne].production = temp;
e[ne].dat = surv_date;
for (i = 0; i < nstages; ++i)
	e[ne].prop[i] = 100.0;
++ne;
clrscr();
newmsg("This section asks for the density or numbers of dung piles found"
	   " on the survey\n\rPlease specify the units that will be used for"
	   " these values    ([Esc] to quit)",FALSE);
if ((units = menu(44,4,1,N_UNITS,units_list,"Units",&m,MENU_FCOL,MENU_BCOL,
				 MENU_HICOL,NULL)) == -1)
	quit();
clearmenu(&m);
cputs("\n\r");
qsort(e,ne,sizeof(e[0]),comp_date);
line = wherey() + 1 + nstages;
for (s = nstages - 1; s >= 0; --s)
	{
	sprintf(str,"Enter estimated %s of %s\n\rfound on the survey (leave blank"
			" if unknown)",(units == 0) ? "number":"density",
			strfirst(units_list[units]));
	newmsg(str,FALSE);
	cprintf("%s : ",list_stages(str,s));
	if (pile_density[s] > 0.0)
		sprintf(str,"%g",pile_density[s] /= units_conversion[units]);
	else
		str[0] = '\0';
	xgets(str,line - s,30,line - s,77,47,xgetch,INP_FCOL,INP_BCOL);
	pile_density[s] = (strlen(str) > 0) ? atof(str) : -1.0;
	f[s] = f_without_prod = var = 0.0;
	for (i = 0; i < ne - 1; ++i)
		{
		if (i >= 1)
			{
			tdiff = (double)e[i + 1].dat - e[i - 1].dat;
			var += e[i].prop[s]*(100.0 - e[i].prop[s])*tdiff*tdiff/
			  e[i].n_original;
			}
		f_without_prod += (temp = (e[i + 1].dat - e[i].dat)*(e[i + 1].prop[s]
		  + e[i].prop[s]));
		f[s] += (e[i].production + e[i + 1].production)*temp;
			/** = av.production * area of trapezium **/
		}
	f[s] *= 0.0025;
	/** Scaling by 1/2 to average production and 1/2 to average prop and by **/
	/**	0.01 to convert from percentages is done outside loop for efficiency **/
	tdiff = (double)e[1].dat - e[0].dat;
	var += e[0].prop[s]*(100.0 - e[0].prop[s])*tdiff*tdiff/e[0].n_original;
	cv[s] = sqrt(var)/f_without_prod;
	cputs("\n\r");
	}
clrscr();
for (i = 0; i < nstages; ++i)
	if (pile_density[i] > 0.0)
		{
		if (units > 1)
			pile_density[i] *= units_conversion[units];
		ele_density[i] = pile_density[i]/f[i];
		}
if ((dcfg = fopen(DCFG_FILE,"w")) != NULL)
	{
	fprintf(dcfg,"[Most recent file used by DUNGSURV] %s\n",filename);
	fprintf(dcfg,"[Defecation rate at time of survey] %g\n",surv_prod);
	for (i = 0; i < nstages; ++i)
		fprintf(dcfg,"[Dung density per km\xFD,stage %s] %g\n",stname[i],pile_density[i]);
	fclose(dcfg);
	}
sprintf(str,"Estimated elephant %s prior to survey on %d/%d/%d",
		(units == 0) ? "numbers":"density per km\xFD",sday,smonth,syear);
puts_at(j = 40 - strlen(str)/2,2,str);
for (i = 0; i < strlen(str); ++i,++j)
	putch_at(j,3,0xCD);
strcpy(str,(units == 0) ? "Est. number ":"Est. density");
cprintf("\n\n\rBased on dung                %s       Calibration      "
		"%s\n\r",str,str);
cputs("piles in                     of dung piles      factor"
	  " (and CV)  of elephants\n\n\r");
for (warnings = 0,i = nstages - 1; i >= 0; --i)
	{
	if (pile_density[i] >= 0.0)
		{
		cprintf("%-26s %15.3f %12.2f (%5.3f) %12.3f",list_stages(str,i),
				pile_density[i],f[i],cv[i],ele_density[i]);
		if (e[0].prop[i] > 1.0)
			{
			++warnings;
			textcolor(FLAGCOL);
			cputs(" *");
			textcolor(FCOL);
			}
		}
	else
		cprintf("%-26s %28.2f (%5.3f)",list_stages(str,i),f[i],cv[i]);
	cputs("\n\r");
	}
if (warnings > 0)
	{
	textcolor(HICOL);
	cprintf("\n\r  Warning: The estimat%s marked with",(warnings == 1) ? "e":
			"es");
	textcolor(FLAGCOL);
	cputs(" * ");
	textcolor(HICOL);
	cprintf("%s biased upwards since some dung\n\r           piles remain"
			" from the first experiment\n\r           (see the \"%% dung"
			" piles surviving to survey date\" graph)",(warnings == 1) ?
			"is":"are");
	textcolor(FCOL);
	}
newmsg("Creating .PLT files...",FALSE);
if (!(dt = (double *) malloc(ne*sizeof(double))))
	gfx = FALSE;
else if (!(y = (double *) malloc(ne*sizeof(double))))
	{
	free(dt);
	gfx = FALSE;
	}
else
	{
	for (i = 0; i < ne; ++i)
		dt[i] = (double) e[i].dat - surv_date;
	open_plotter_file(PLOTFILE);
	initialise_plotter();
	set_user_coords((int)(1.05*dt[0]),(int)(-0.05*dt[0]),-5,105);
	pen(1);
	if (dt[0] > -14.0)
		xstep = 1.0;
	else
		for (xstep = 7.0; dt[0]/xstep < -14.0; xstep *= 2.0);
	sprintf(str,"Date fresh dung piles observed (number of days before %u/%u/%u)",
			sday,smonth,syear);
	axes(str,0.0,dt[0],0.0,xstep*((int)dt[0]/(int)xstep),xstep,1.0,0,TRUE,
		 "% remaining by survey date",dt[0],0.0,100.0,0.0,25.0,1.0,0,FALSE);
	char_size(A4CHARWIDTH*0.8,A4CHARHEIGHT*0.8);
	for (i = 0; i < nstages; ++i)
		{
		for (j = 0; j < ne; ++j)
			y[j] = e[j].prop[i];
		pen(i + 2);
		poly_line(ne,dt,y);
		for (j = 1; y[j] == 0.0 && j < ne; ++j);
		list_stages(str,i);
		label(dt[j],y[j],str,0.1,(i == 0) ? 0.5 : (-0.5 - strlen(str)));
		}
	set_line_type(1,2.0);
	pen(7);
	for (i = 0; i < ne; ++i)
		draw_line(dt[i],0.0,dt[i],y[i]);
	pen(0);
	open_plotter_file(WEIGHTFILE);
	maxwt = 0.0;
	for (s = 0; s < nstages; ++s)
		for (i = 0; i < ne; ++i)
			if ((e[i].prop[s] *= e[i].production/f[s]) > maxwt)
				maxwt = e[i].prop[s];
	for (wfact = 1.0,wdp = -1; maxwt < 100.0; wfact *= 10.0,maxwt *= 10.0,
		 ++wdp);
	if (wdp < 0)
		wdp = 0;
	if (wfact > 1.0)
		for (i = 0; i < ne; ++i)
			for (s = 0; s < nstages; ++s)
				e[i].prop[s] *= wfact;
	initialise_plotter();
	set_user_coords(1.05*dt[0],-0.3*dt[0],-0.05*maxwt - 0.5,1.05*maxwt + 0.5);
	pen(1);
	for (ystep = 1.0; maxwt/ystep > 6.0; ystep *= 2.0);
	sprintf(str,"Number of days before %u/%u/%u",sday,smonth,syear);
	axes(str,0.0,dt[0],0.0,xstep*((int)dt[0]/(int)xstep),xstep,1.0,0,TRUE,
		 "% weighting",dt[0],0.0,maxwt,0.0,maxwt*0.2,1.0/wfact,wdp,FALSE);
	char_size(A4CHARWIDTH*0.8,A4CHARHEIGHT*0.8);
	for (i = 0; i < nstages; ++i)
		{
		for (j = 0; j < ne; ++j)
			y[j] = e[j].prop[i];
		pen(i + 2);
		poly_line(ne,dt,y);
		for (j = 1; y[j] == 0.0 && j < ne; ++j);
		list_stages(str,i);
		label(dt[ne - 1],y[ne - 1],str,-0.5,0.5);
		}
	pen(0);
	close_plotter_file();
	free(y);
	free(dt);
	}
if (gfx)
	{
	newmsg("Select a graph to view on the screen  or press [Esc] to exit",FALSE);
	gettext(1,1,80,25,scrnbuff);
	while((j = menu(20,((warnings > 0) ? 22 : 19) - N_PLOT,1,N_PLOT,
		  plot_choices,"Graphs",&m,MENU_FCOL,MENU_BCOL,MENU_HICOL,NULL)) != -1)
		{
		clearmenu(&m);
		if (j == 0)
			seeplot(PLOTFILE);
		else if (j == 1)
			seeplot(WEIGHTFILE);
		getch();
		restorecrtmode();
		puttext(1,1,80,25,scrnbuff);
		}
	}
else
	{
	newmsg("The graphs cannot be displayed on this machine\n\rPlease press"
		   " any key when ready to exit",FALSE);
	getch();
	}
quit();
}

static unsigned days(const char *s,int *day,int *month,int *year)
{
static unsigned cumdays[12]={0,31,59,90,120,151,181,212,243,273,304,334};
int y;
unsigned d;
char *scopy;

scopy = (char *) malloc(strlen(s) + 1);
strcpy(scopy,s);
*day = atoi(strtok(scopy," /.-"));
*month = atoi(strtok(NULL," /.-"));
*year = atoi(strtok(NULL," /.-"));
free(scopy);
if (!check_date(*day,*month,year))
	return(0);
if ((y = *year) < 1980)
	return(0);
d = *day + cumdays[*month - 1] + 365*(y - 1980);
if (*month < 3)
	--y;
if (y >= 1980)
	d += 1 + (y - 1980)/4;
if (y >= 2000)
	--d;
return(d);
}

static void newmsg(char text[],char highlight)
{
TEXT_STATE ts;

savetextstate(&ts);
window(1,1,80,2);
if (highlight)
	{
	beep_at(880,50);
	textcolor(WHITE);
	textbackground(RED);
	}
else
	{
	textcolor(BLACK);
	textbackground(LIGHTGRAY);
	}
clrscr();
puts_at(1,1,text);
restoretextstate(&ts);
}

static void quit(void)
{
if (dcfg)
	fclose(dcfg);
newmsg("DUNGSURV program completed",FALSE);
textattr(NORMAL);
_wscroll = 0;
window(1,24,80,25);
clrscr();
gotoxy(1,1);
exit(0);
}

static int comp_date(const void *a,const void *b)
{
experiment_data *x,*y;

x = (experiment_data *) a;
y = (experiment_data *) b;
if (x->dat == y->dat)
	return(0);
return ((x->dat > y->dat) ? 1 : -1);
}

static char *list_stages(char *s,int upto)
{
int i;

if (upto == 0 && nstages > 1)
	sprintf(s,"Stage %s only",stname[0]);
else if (upto == 1 && nstages > 2)
	sprintf(s,"Stages %s and %s only",stname[0],stname[1]);
else if (upto == 2 && nstages > 3)
	sprintf(s,"Stages %s, %s and %s only",stname[0],stname[1],stname[2]);
else
	{
	strcpy(s,"All stages");
	if (upto < nstages - 1)
		{
		strcat(strcat(s," except "),stname[upto + 1]);
		for (i = upto + 2; i < nstages - 1; ++i)
			strcat(strcat(s,", "),stname[i]);
		if (upto != nstages - 2)
			strcat(strcat(s," and "),stname[nstages - 1]);
		}
	}
return(s);
}

static char *read_line(char *s)
{
return(strip(fgets(s,132,input),"[]"));
}
